// Transfer data between P6 and PC
//
// compile Mac: gcc -DMACOSX -lftd2xx -o transfer transfer.c
// compile Win: gcc -DWINDOWS -o transfer.exe transfer.c FTD2XX.lib
//

#include <stdio.h>
#ifdef MACOSX
#include "WinTypes.h"
#else
#include <windows.h>
#endif
#include "ftd2xx.h"

#define OK	1
#define	NG	2

// commands
#define LOAD_DATA	1	// load data
#define SAVE_DATA	2	// save data

// to exit
static int exitNow = 0;
static void OnBreak(int Arg) { exitNow = 1; signal(Arg,OnBreak); }

// initialize USB controller
static int initUSB(FT_HANDLE *ftHandle)
{
	DWORD bytesWritten;
	FT_STATUS ftStatus;

	// Open
	ftStatus = FT_Open(0, ftHandle);
	if (ftStatus != FT_OK) {
		fprintf(stderr, "Can't initialize USB controller.\n");
		return 1;
	}

	// Reset device
	ftStatus = FT_ResetDevice(*ftHandle);
	if (ftStatus != FT_OK) {
		fprintf(stderr, "Can't reset USB controller.\n");
		return 1;
	}

	// Set latency
	ftStatus = FT_SetLatencyTimer(*ftHandle, 1);
	if (ftStatus != FT_OK) {
		fprintf(stderr, "Can't setup USB controller.\n");
		return 1;
	}
 	return 0;
}

// output to USB controller
static int outUSB(FT_HANDLE ftHandle, BYTE *buff, DWORD size)
{
	DWORD bytesWritten;
	FT_STATUS ftStatus;

	ftStatus = FT_Write(ftHandle, buff, size, &bytesWritten);
	if (ftStatus != FT_OK || bytesWritten != size) {
		fprintf(stderr, "Can't output to USB controller.\n");
		return 1;	
	}
	return 0;
}

// input from USB controller
static int inUSB(FT_HANDLE ftHandle, BYTE *buff, DWORD size)
{
	DWORD bytesRead;
	FT_STATUS ftStatus;

	ftStatus = FT_Read(ftHandle, buff, size, &bytesRead);
	if (ftStatus != FT_OK || bytesRead != size) {
		fprintf(stderr, "Can't input from USB controller.\n");
		return 1;
	}
	return 0;
}

static void ReadFileWord(FILE *fp, WORD *val)
{
	*val = fgetc(fp);
	*val += (fgetc(fp)*256);
}

static void SendWord(FT_HANDLE ftHandle, WORD val)
{
	BYTE buff[2];

	buff[0] = (val & 0xff);
	buff[1] = ((val>>8) & 0xff);
	outUSB(ftHandle, buff, 2);
}

static void ReceiveWord(FT_HANDLE ftHandle, WORD* val)
{
	BYTE buff[2];

	inUSB(ftHandle, buff, 2);
	*val = (WORD)buff[0]+(WORD)buff[1]*256;
}

// load program to P6
static int LoadProgram(FT_HANDLE ftHandle, char *file)
{
	FILE *fp;
	BYTE buff[4096];
	WORD beginAdr, endAdr, startAdr;
	DWORD i;
	size_t read;

	printf("Loading Program \"%s\"\n", file);
	if (fp = fopen(file, "rb")) {

		// begin address
		ReadFileWord(fp, &beginAdr);
		SendWord(ftHandle, beginAdr);

		// end address
		ReadFileWord(fp, &endAdr);
		SendWord(ftHandle, endAdr);

		// start address
		ReadFileWord(fp, &startAdr);
		SendWord(ftHandle, startAdr);

		for (i=beginAdr; i<=endAdr;) {
			if (endAdr-i+1 > 4096) {
				read = fread(buff, 1, 4096, fp); 
				outUSB(ftHandle, buff, read);
			} else {
				read = fread(buff, 1, endAdr-i+1, fp); 
				outUSB(ftHandle, buff, read);
			}
			i += read;
		}
		fclose(fp);
		printf("OK\n");
		return 0;
	}
	printf("NG\n");
	return 1;
}

// load data to P6
static int LoadData(FT_HANDLE ftHandle, char *file)
{
	FILE *fp;
	BYTE buff[4096];
	WORD length;
	DWORD i;
	size_t read;

	printf("Loading Data \"%s\"\n", file);
	if (fp = fopen(file, "rb")) {
		// receive length
		ReceiveWord(ftHandle, &length);

		// reaction 
		buff[0] = OK;
		outUSB(ftHandle, buff, 1);

		for (i=0; i<length;) {
			if (length-i > 4096) {
				read = fread(buff, 1, 4096, fp); 
				outUSB(ftHandle, buff, read);
			} else {
				read = fread(buff, 1, length-i, fp); 
				outUSB(ftHandle, buff, read);
			}
			i += read;
		}
		fclose(fp);
		printf("OK\n");
		return 0;
	}
	printf("NG\n");
	return 1;
}

// save data from P6
static int SaveData(FT_HANDLE ftHandle, char *file)
{
	FILE *fp;
	BYTE buff[4096];
	WORD length;
	DWORD i;
	size_t read;

	printf("Saving Data \"%s\"\n", file);
	if (fp = fopen(file, "wb")) {
		// receive length
		ReceiveWord(ftHandle, &length);

		// reaction 
		buff[0] = OK;
		outUSB(ftHandle, buff, 1);

		for (i=0; i<length;) {
			if (length-i >= 4096) {
				inUSB(ftHandle, buff, 4096);
				fwrite(buff, 1, 4096, fp);
				i += 4096;
			} else {
				inUSB(ftHandle, buff, length-i);
				fwrite(buff, 1, length-i, fp);
				i = length;
			}
		}
		fclose(fp);
		printf("OK\n");
		return 0;
	}
	printf("NG\n");
	return 1;
}

int main(int argc, char *argv[])
{
	FT_HANDLE ftHandle;
	BYTE file[7];
	BYTE ch;

	if (argc < 2) return 1;

	// init USB controller
	if (initUSB(&ftHandle)) return 1;

	// load main program
	LoadProgram(ftHandle, argv[1]);

	// waiting command
	do {
		inUSB(ftHandle, &ch, 1);
		switch (ch) {
		case LOAD_DATA: // load data
			inUSB(ftHandle, file, 7);
			LoadData(ftHandle, file);
			break;
		case SAVE_DATA: // save data
			inUSB(ftHandle, file, 7);
			SaveData(ftHandle, file);
			break;
		default:
			fprintf(stderr, "Unknown command %c\n",ch);
			break;
		}
	} while (!exitNow);

	// close USB controller
	FT_Close(ftHandle);
	return 0;
}
